/** author Tim vor der Brck 2000
 * this software is released under the GNU General Public License Version 2
 */


#include <stdexcept>
#include "StringParser.h"

StringParser::StringParser(string tokStr,string delim[],int numDelim,bool* include,int startPos): 
  m_tokStr(tokStr),m_numDelim(numDelim),m_include(include),m_startPos(startPos)
{
  // clone delimiter array
  m_delim=new string[m_numDelim];
  for (int i=0;i<m_numDelim;i++)
    m_delim[i]=delim[i];
  m_endIter=new StringIterator(this,-1);
}

StringParser::~StringParser()
{
  delete[] m_delim;
  delete m_endIter;
}

StringParser::const_iterator& StringParser::begin()
{
  StringIterator* iter=new StringIterator(this,m_startPos);
  iter->m_nextToken=iter->findNextToken();
  if (iter->m_nextToken=="") iter->m_pos=-1;
  return *iter;
}

  /** returns the iterator that points after the last element*/
StringParser::const_iterator& StringParser::end()
{
  return *m_endIter;
}


StringIterator::StringIterator(StringParser* parser,int pos): m_parser(parser),m_pos(pos)
{
  if (m_pos==-1) m_nextToken=""; //iterator points to end of string
}

StringIterator::StringIterator(StringIterator& iter): m_parser(iter.m_parser),m_pos(iter.m_pos),
						      m_nextToken(iter.m_nextToken)
{
}

StringIterator& StringIterator::operator++()
{
  if (*this==m_parser->end()) return *this;
  m_nextToken=findNextToken();
  if (m_nextToken=="") m_pos=-1; //return end
  return *this;
}

string StringIterator::operator*()
{
  if (m_pos==-1 && m_nextToken=="") throw out_of_range(string("iterator points after the last element"));
  return m_nextToken;
}

/** return "" if there is no next token*/
string StringIterator::findNextToken()
{
  string nextToken="";
  bool findEnd=false,isDelim; //findEnd find the end of a token, otherwise find beginning
  int i;
  while (m_pos<(m_parser->m_tokStr).length())
  {
      isDelim=false;
      for (i=0;i<(m_parser->m_numDelim);i++)
      {
	//delimiter at current position?
	if (m_pos+(m_parser->m_delim[i]).length()-1<(m_parser->m_tokStr).length() && 
	    (m_parser->m_tokStr).substr(m_pos,(m_parser->m_delim)[i].length())==(m_parser->m_delim)[i])
	  /* Fuzzy.equals(m_parser.m_tokStr.substring(m_pos,m_pos+m_parser.m_delim[i].length()),m_parser.m_delim[i],
	     m_ignoreCase,m_fuzzy))*/
	{ //delimiter?
	  if ((m_parser->m_include)!=0l && (m_parser->m_include)[i]) //include delimiter in token?
	  {
	    
	    if (nextToken!="") return nextToken; //???
	    //if (!nextToken.equals("")) return nextToken;
	    nextToken+=(m_parser->m_delim)[i];
	    findEnd=true;
	    m_pos+=(m_parser->m_delim)[i].length();
	    isDelim=true;
	    break;
	  }
	  else
	  {
	    m_pos+=(m_parser->m_delim)[i].length();
	    if (findEnd) return nextToken;
	    isDelim=true;
	    break;
	  }
	}
      } //for (i=0;i<m_parser.m_delim.length;i++)
      
      if (!isDelim) //no delimiter construct token
      {
	//nextToken+=(new Character(m_parser.m_tokStr.charAt(m_pos++))).toString();
	nextToken+=string(1,m_parser->m_tokStr[m_pos++]);
	findEnd=true; //find now end of token
      }
  }  //while (m_pos<(m_parser->m_tokStr).length())
  return nextToken;
}

bool StringIterator::operator==(StringIterator& iter)
{
  return (m_parser==iter.m_parser && m_pos==iter.m_pos && m_nextToken==iter.m_nextToken);
}

bool StringIterator::operator!=(StringIterator& iter)
{
  return (!(*this==iter));
}

