#ifndef boardviewh
#define boardviewh

#include "icon.h"
#include "counter.h"

/* copyright by Tim vor der Brck, distributed under the GNU General Public License Version 2
 * represents the graphical chess board */ 

class CBoard
{
public:
  virtual int getcontents(int x,int y)=0;
};

class CGame
{
public:
  /* returns the icon of one counter */
  virtual CIcon* convvaluetoicon(int value)=0;

  /* draws an empty board */
  virtual void drawboard(Display* display,GC gcbg,Pixmap& pixmap,int dx,int dy)=0;
};

class CBoardview : public CPlayfield
{
 public:
  CBoardview();
  ~CBoardview(); 

   /* should be called first and only once */
  void create(CGame* g,CBoard* board,Display* display,Window win,int width,int height,
              int rows,int columns);

  /* removes an icon at the specified position */
  void removeicon(int x,int y,CIcon* icon);

  /* draws an icon at the specified position */
  void drawicon(int x,int y,CIcon* icon);

  /* returns the contents of field x,y */
  CCounter* getcontents(int x,int y);
 
  /* set contents of field x,y with counter ss */
  void setcontents(int x,int y,CCounter* ss);

  /* resizes the window */
  void resize(int width,int height);
 
  /* changes the dimensions of the board */
  void changedimensions(int columns,int rows);

  /* sets the output window */
  void setwindow(Window window);

  /* draws a part of the board */ 
  void draw(int x,int y,int width,int height);
  bool begindrag(int winx,int winy,int x,int y,bool bmove);
  void dodrag(int winx,int winy);
  void drop(int winx,int winy,int& x,int& y);

  /* updates the look of the board according to the logical board*/
  void update();

  /* sets all fields to NULL and erases any NOT NULL entrys */
  void eraseboard();

  /* draws the whole board */
  void draw();

  /* redraws the board according to its visual discription,
     independent of the logical board description, calls
     drawemptyboard first */
  void redraw();

 /* converts a board to a window position */
  int convxtowinx(int x,bool centre,int width);

  int convytowiny(int y,bool centre,int height);
  int convwinxtox(int winx,bool centre,int width);
  int convwinytoy(int winy,bool centre,int height);

 protected:

  /* sets how many couloms and rows the board has */ 
  void setdimensions(int columns,int rows);

 

  /* sets all fields to NULL */
  void deleteboard();

  /* draws an empty board in the pixmap background, double buffers */
  void drawemptyboard();

  /* allocates new pixmaps and eventually deletes the old,
     needed by creation and resizing */
  void setuppixmaps();
  
  Pixmap m_pmemptyboard;
  Pixmap m_pmboard;
  Pixmap m_pmdnd;
  Window m_win;
  Window m_root;
  Display* m_display;
  int m_depth;
  int m_screennr;
  Window m_defroot;
  int m_fdx;
  int m_fdy;
  int m_dx;
  int m_dy;
  CCounter*** m_contents;
  int m_rows;
  int m_columns;
  GC m_gc;
  bool m_bonscreen; //draw on screen or only in the dubble buffer
  CGame* m_game;
  CBoard* m_board;
  CIcon* m_dragicon;
  int m_hotx;
  int m_hoty; 
};


#endif



