/*Copyright: Tim vor der Brck 1999

This software is released under the GNU General Public License.
It may be redistributed as long this copyright remark is not
removed or modified.
*/


/* csbrett.cpp*/
#include <math.h>
#include <strstream.h>
#include <stdlib.h>
#include <time.h>
#include <iostream.h>
#include "csbrett.h"
 

const int dxa[]={0,1,0,-1};
const int dya[]={1,0,-1,0};

/*************** operator<< fr Zug ***********/
ostream& operator<<(ostream& os,const zug& z)
{
  if (z!=leererzug)
    os << char(z.x1+'A') << z.y1+1 << '-'
       << char(z.x2+'A') << z.y2+1;
  return os;
}

/**************** operator>> fr Zug************/
istream& operator>>(istream& is,zug& z)
{
  char c,c1,c2;

  is >> c1 >> z.y1 >> c >> c2 >> z.y2;
  z.x1=int(c1-'A');
  z.x2=int(c2-'A');
  z.y1--;
  z.y2--;
  return is;
}




/******************* BERALPHABETA ******************/
void beralphabeta(const int&i,int& alpha,int& beta)
{
  alpha=((i-1)/4+1)*2-1;
  beta=alpha+((i-1)%4)*2;
}


/*************** Desktruktor *******************/
csbrettklasse::~csbrettklasse()
{
	FreeMatrix(ind,anztiefen);
	FreeMatrix(bew,anztiefen);
	FreeMatrix(zi,anztiefen);
}

//*********************** INIT ********************************
void csbrettklasse::init() {

//srand(time(NULL));
hz=0;
verbzug=leererzug;
kaiserx[0]=4;
kaisery[0]=0;
kaiserx[1]=4;
kaisery[1]=ymax;
for (int x=0;x<=xmax ;x++ ) {
  for (int y=1;y<ymax ;y++ ) f[x][y]=leer;
}
f[4][0]=kaiser;
f[4][ymax]=-kaiser;
f[3][0]=f[5][0]=mandarin;
f[3][ymax]=f[5][ymax]=-mandarin;
f[2][0]=f[6][0]=elefant;
f[2][ymax]=f[6][ymax]=-elefant;
f[1][0]=f[7][0]=pferd;
f[1][ymax]=f[7][ymax]=-pferd;
f[0][0]=f[8][0]=turm;
f[0][ymax]=f[8][ymax]=-turm;
f[1][2]=f[7][2]=geschuetz;
f[1][ymax-2]=f[7][ymax-2]=-geschuetz;
f[0][3]=f[2][3]=f[4][3]=f[6][3]=f[8][3]=bauer;
f[0][6]=f[2][6]=f[4][6]=f[6][6]=f[8][6]=-bauer;
}

/***************** MOEGLICHVONFIGUR **********************/
bool csbrettklasse::moeglichvonfigur(const zug& z){
int x,y,dx=z.x2-z.x1,dy=z.y2-z.y1,sx=sgn(dx),sy=sgn(dy);
bool sprung=false;
switch (abs(f[z.x1][z.y1])) {
 case kaiser:
   if (!impalast(z.x2,z.y2)) return false;
   return (abs(dx)==1 && dy==0 || dx==0 && abs(dy)==1);
 case turm:
   if (dx!=0 && dy!=0) return false;
   for (x=z.x1+sx,y=z.y1+sy;x!=z.x2 || y!=z.y2;x=x+sx,y=y+sy)
     if (f[x][y]!=leer) return false;
   return true;
 case geschuetz:
   {
     if (dx!=0 && dy!=0) return false;
     
     bool occupiedField=false; //was there an occupied field alreay
     for (x=z.x1+sx,y=z.y1+sy;x!=z.x2 || y!=z.y2;x=x+sx,y=y+sy)
     {
       if (f[x][y]!=leer)
       {
	 if (occupiedField) return false; //2 occupied fields between start end end
	 occupiedField=true;
       }
     }
     if (occupiedField && f[z.x2][z.y2]==leer) //cannot move on empty field with a piece between
       return false;
     if (!occupiedField && f[z.x2][z.y2]!=leer) //cannot move on occupied field if no piece between
       return false;
     return true;
   }
 case pferd:
   if (!(abs(dx)==2 && abs(dy)==1 || abs(dx)==1 && abs(dy)==2))
     return false; //1 grad, eins schrg
   return (f[z.x2-sx][z.y2-sy]==leer); //Vorderfeld leer
 case elefant:
   if (f[z.x1][z.y1]>0 && z.y2>4) return false;
   if (f[z.x1][z.y1]<0 && z.y2<5) return false;
   if (abs(dx)!=2 || abs(dy)!=2) return false; //2 Felder diagonal
   return (f[z.x1+sx][z.y1+sy]==leer); //Zwischenfeld leer
 case bauer:
   if (dx==0 && dy==sgn(f[z.x1][z.y1])) return true; //1 Feld n.vorne
   if (sgn(f[z.x1][z.y1])==1 && z.y2>4 && abs(dx)==1 && dy==0)
      return true; //roter Bauer zur Seite
   return (sgn(f[z.x1][z.y1])==-1 && z.y2<5 && abs(dx)==1 && dy==0);
       //schwarzer Bauer zur Seite
 case mandarin:
   return (impalast(z.x2,z.y2) && abs(dx)==1 && abs(dy)==1);
} //endeswitch
return 0;
} //endemoeglichvonfigur

/************************ SCHACH **************************/
bool csbrettklasse::schach(const int& sp) {
/* berprft ob Spieler sp im Schach ist*/
zug z;
stein h;
z.x2=kaiserx[sp];
z.y2=kaisery[sp];

for (z.x1=0;z.x1<=xmax;z.x1++) {
  for (z.y1=0;z.y1<=ymax;z.y1++) {
    h=abs(f[z.x1][z.y1]);
    if (h!=leer && spvonstein(f[z.x1][z.y1])!=sp) {//!eig.Sp.
      if (h!=elefant && h!=mandarin && h!=kaiser) {
        if (h!=bauer || abs(z.x2-z.x1)==1 || abs(z.y2-z.y1)==1) {
          if (moeglichvonfigur(z)) return true;
        }
      }
    }
  }
}
return false;
}

//************************* REMIS ***************************
bool csbrettklasse::remis() {
for (int x=0;x<=xmax ;x++ ){
  for (int y=0;y<=ymax ;y++ ){
    stein h=abs(f[x][y]);
    if (h!=leer && h!=mandarin && h!=kaiser && h!=elefant)
       return false;
  }
}
return true;
}

//******************** ALLGMOEGLICH *************************
bool csbrettklasse::allgmoeglich(const zug& z)
{
/* berprft ab ein bei einem Zug kein eigener Stein auf dem
Zielfeld steht und ob die beiden Kaiser nicht in einer Linie stehen
ohne, da ein Stein dazwischen steht*/

  bool rc=true;
  stein h;

  //auf Zielfeld steht eigener Stein:
  if (sgn(f[z.x1][z.y1])==sgn(f[z.x2][z.y2])) return false;

  //knnen Kaiser sich sehen?
  if (z.x1<3 || z.x1>5) return true;
  h=ziehe(z); //direkte Verbindung zwischen Kaiser?
  if (kaiserx[0]!=-1 && kaiserx[1]!=-1 && kaiserx[0]==kaiserx[1])
  {
    for (register int y=kaisery[0]+1;y<kaisery[1];y++ )
    {
      if (f[kaiserx[0]][y]!=leer) break;
      if (y==kaisery[1]-1) rc=false;
    }
  }
  zuruecknehmen(z,h); //den Kaisern?
  return rc;
}


//************************ ZUEGEVONSPIELER ***********************
int csbrettklasse::zuegevonspieler
      (const int& sp,long* b0,bool schrbew,bool vollesuche,
       const int&t,int *ix){
int x,y,k=0;
for (x=0;x<=xmax ;x++ ) {
  for (y=0;y<=ymax ;y++ ) {
    if (spvonstein(f[x][y])==sp && f[x][y]!=leer) {
         //Feld ist von Spieler sp besetzt?
      if(zuegevonfeld(x,y,t,k,b0,schrbew,vollesuche,ix)==false)
      {
        zi[t][0]=leererzug; //Mensch-Optimierung durchgefhrt
        return -1;
      }
    } /* endif */
  } /* endfor */
} /* endfor */
zi[t][k]=leererzug;
return k;
}

//***************************** ANALYSE ********************
/* analysiert die Spielsituation 
sp sollte dem Spieler entsprechen, der gerade gezogen hat*/
zctyp csbrettklasse::analyse(const int& sp) {
int anz,i;
bool pattmatt=true;
stein h;
if (remis()) return istremis;
anz=zuegevonspieler(!sp,NULL,false,true,0,NULL);

for (i=0;i<anz;i++) {
  h=ziehe(zi[0][i]);
  if (!schach(!sp)) {
    pattmatt=false;
    zuruecknehmen(zi[0][i],h);
    break;
  }
  zuruecknehmen(zi[0][i],h);
}
if (pattmatt) {
  if (schach(!sp)) return matt;
  return patt;
}
else {
  if (hz==hzmax-1) return istremis;
  if (schach(!sp)) return istschach;
}
return normal;
}

/******************* besterzug ************************/
zug csbrettklasse::besterzug
    (int sp,int& command,int lev)
{
  long b0[60];
  register int l;
  int t=0,l2,index,k,k0,i[60]; //i:Wo bin ich bei Tiefe t
  zug az=leererzug,bestzug=leererzug;
  stein hi[60],zfi; //Speichere geschlagene Figuren,Zielfeldinhalt
  int startsp=sp; //Spieler an der Reihe
  bool opt;


  // Initialisierung
  if (lev==0) lev=level;
  beralphabeta(lev,alpha0,beta0);
  alpha=alpha0; //Alpha-Suchtiefe auf Standardwert
  beta=beta0;  //Beta-Suchtiefe auf Standardwert
  b0[0]=-unendl;
  i[0]=0;

  //Fehler?
  if (lev==0) return leererzug;
  if (beta+3>anztiefen) return leererzug;

  k0=zuegevonspieler(sp,NULL,true,true,0,i);
  bew[0][time(NULL)%k0]+=5+(t<5)*80;

  ordnen(k0,ind[0],bew[0]);
  //Zge von Spieler sp, keine Bewertungsoptimierung
  //Bewertungen speichern,alle Zge durchrechnen

  for (l=1;l<anztiefen;l=l+2) //Fr alle ungeraden Indizes
    for (l2=0;l2<=mzugm;l2++) ind[l][l2]=l2;

  for (;;)
  {
    index=ind[t][i[t]];
    az=zi[t][index]; //aktueller Zug
    zfi=abs(f[az.x2][az.y2]); //Zielfeldinhalt
    if (az!=leererzug && t<beta && zfi!=kaiser)
    //Eins hher
    {
      if (command>=100) //Soll Zug abgebrochen werden
      {
	    for (l=t-1;l>=0;l--)
          zuruecknehmen(zi[l][ind[l][i[l]]],hi[l]);
        return leererzug;
      }

      /* Computeroptimierung
         Man ermittle den letzten Zug des Computers.
         Ist dort eine Bewertung niedriger als eine fertige
         Bewertung wei man, da dieser Zug nur noch schlechter
         werden kann, weil im nchsten Zug der Mensch zum
         letzten Mal zieht*/
      opt=false;
      if (t==beta-1 || t==alpha-1 && b0[t]>0)
      {
        for (l=0;l<=t;l=l+2)
        {
          if (bew[t][index]<=b0[l])
          {
            opt=true;
            break;
          }
        }
      }
      if (opt) zi[t][index]=leererzug;
      else
      {
        hi[t]=ziehe(zi[t][index]); //Zug durchfhren
        t++; //Tiefe erhhen
        i[t]=0; //Wieder von vorne anfangen
        sp=!sp;
        k=zuegevonspieler(sp,b0,true,t<=alpha,t,i);
        // Zge von Spieler sp,
        // Bewertungen speichern, evtl. alle Zge ermitteln

	// Zugmglichkeitsstrategie
      //  if (k>0 && k<40) bew[t-1][ind[t-1][i[t-1]]]+=40-k;

        if (k==-1) b0[t]=unendl; //Mensch-Optimierung
        else if (t>alpha && !k) b0[t]=0; //da nicht alle Zge gepr.
        else b0[t]=-unendl;

        
        //evtl. Spielstrke erhhen
        if (t==1 && i[0]==0)
        {
          if (k0*k<=30 && k>0)
          {
            if (alpha0!=beta0) alpha=beta0;
            else
            {
              beta=beta0+2;
              alpha=beta;
            }
          }
          else if (k0*k<=100) beralphabeta(level+1,alpha,beta);
	  }

        if (sp==startsp) ordnen(k,ind[t],bew[t]);
        //Bewertungen ordnen, wenn Computer an der Reihe

      }
    }


    else if (az==leererzug || zfi==kaiser) //1 down
    {


      if (t==0) 
      {
         if (zfi==kaiser && bestzug==leererzug)
           return az;
        break; //Berechnung zuende
      }
      if (zfi==kaiser) b0[t]=bew[t][index]; //Kaiser geschlagen
      if (t==alpha+1 && b0[t]<-40) b0[t]=-40; //Bew.unsicher

      t--;
      sp=!sp;
      index=ind[t][i[t]];
      bew[t][index]=bew[t][index]-b0[t+1]; //Bewertung vervollst.

      if (bew[t][index]>b0[t]) //7/6
      {
        b0[t]=bew[t][index];
        if (t==0) bestzug=zi[0][index];
      }
      zuruecknehmen(zi[t][index],hi[t]);
      i[t]++; //Nchster Zug an der Reihe
    }
    else //t==beta
    {
      for (l=0;l<k;l++)
        if (bew[t][l]>b0[t]) b0[t]=bew[t][l];
      i[t]=l;
    }
  }
  //if (t==0) cout << b0[0] << endl;
  return bestzug;
}

/****************** AENDERELEVEL *************************/
bool csbrettklasse::setzelevel(int v)
{
	int alpha2,beta2,anztiefen2;
	zug **zi2;
	int **ind2;
	long **bew2;
		
	beralphabeta(v,alpha2,beta2);
	anztiefen2=beta2+3;
	
	//Neuen Speicher allozieren
	AllocMatrix(zi2,anztiefen2,mzugm+1);
	AllocMatrix(ind2,anztiefen2,mzugm+1);
	AllocMatrix(bew2,anztiefen2,mzugm+1);

	//Fehler?
	if (zi2==NULL || ind2==NULL || bew2==NULL)
	{
		FreeMatrix(zi2,anztiefen2);
		FreeMatrix(ind2,anztiefen2);
		FreeMatrix(bew2,anztiefen2);
		return false;
	}

	FreeMatrix(zi,anztiefen);
	FreeMatrix(ind,anztiefen);
	FreeMatrix(bew,anztiefen);
	zi=zi2;
	ind=ind2;
	bew=bew2;
	alpha0=alpha=alpha2;
	beta0=beta=beta2;
	anztiefen=anztiefen2;
	level=v;
	return true;
}


/********************* AKTVERBZUG *****************/
void csbrettklasse::aktverbzug()
{
  verbzug=leererzug;
  if (hz<11) return;

  if (zr[hz-1]!=zr[hz-5] || zr[hz-5]!=zr[hz-9]) return;
  if (zr[hz-2]!=zr[hz-6] || zr[hz-6]!=zr[hz-10]) return;
  if (zr[hz-3]!=zr[hz-7] || zr[hz-7]!=zr[hz-11]) return;
  if (zr[hz-4]!=zr[hz-8]) return;
  for (int i=hz-11;i>=hz-8;i++)
    if (h[i]!=0) return;
  verbzug=zr[hz-4];
}

/***************** ZIEHEREG ****************/
bool csbrettklasse::ziehereg(const zug& z)
{
  if (hz>hzmax) return false;
  stmoved[hz]=f[z.x1][z.y1];
  h[hz]=ziehe(z); //Zug auf Brett durchfhren
  zr[hz]=z; //Zug abspeichern
  hz++; //Nchster Halbzug
  aktverbzug();
  return true;
}

/******************* ZURUECKNEHMENREG ***************/
stein csbrettklasse::zuruecknehmenreg()
{
  if (hz==0) return 100;
  zuruecknehmen(zr[hz-1],h[hz-1]);
  hz--;
  aktverbzug();
  return h[hz];
}

/******************* SETZEBRETT *********************/
void csbrettklasse::setzebrett(stein feld2[xmax+1][ymax+1])
{
	hz=0; //Spielanfang
	verbzug=leererzug;
	for (int x=0;x<=xmax;x++)
	{
		for (int y=0;y<=ymax;y++)
		{
			f[x][y]=feld2[x][y]; //Feld kopieren
			if (f[x][y]==kaiser) //Position des Kaisers setzen
			{
				kaiserx[0]=x;
				kaisery[0]=y;
			}
			else if (f[x][y]==-kaiser)
			{
				kaiserx[1]=x;
				kaisery[1]=y;
			}
		}
	}
}

//********************* operator << fuer brett *******************
ostream& operator<<(ostream& os,const csbrettklasse& b)
{
  int x,y,i;
  os << b.hz << endl;

  for (i=0;i<b.hz;i++)
    os << b.zr[i] << endl;

  for (x=0;x<=xmax;x++)
  {
    for (y=0;y<=ymax;y++)
    {
      os << b.f[x][y] << '\t';
    } 
    os << '\n';
  }

  for (i=0;i<b.hz;i++)
      os << b.stmoved[i] << endl;
}

//*************** operator >> fuer brett ***********************
istream& operator>>(istream& is,csbrettklasse& b)
{
  int x,y,i;
 
  is >> b.hz;

  for (i=0;i<b.hz;i++)
  {
    is >> b.zr[i];
  }

  for (x=0;x<=xmax;x++)
    for (y=0;y<=ymax;y++)
      is >> b.f[x][y];

  for (i=0;i<b.hz;i++)
  {
      is >> b.stmoved[i];
  }
}



zug csbrettklasse::holezug(int i)
{
	if (i<0 || i>hz) return leererzug;
	return zr[i];
}


