
/*Copyright: Tim vor der Brueck 1999

This software is released under the GNU General Public License.
It may be redistributed as long this copyright remark is not
removed or modified.
*/
/* Brett.h */

#ifndef csbretthh
#define csbretthh
#include <math.h>
#include <iostream.h>
#include <stdlib.h>
#include <vbase/array.h>
#include <vmath/VMath.h>
#include <vgame/boardview.h>


#define impalast(x,y) (2<x && 6>x && y>=0 && y<=ymax && \
                      (y<3 || y>ymax-3))
#define imfeld(x,y) (0<=x && x<=xmax && 0<=y && y<=ymax)
#define spvonstein(x) (x<0)
//liefert Spieler (0 oder 1) von Stein

typedef struct {int x1; int y1; int x2; int y2;} zug;
typedef int stein;
typedef int bctyp;
typedef enum {normal,istschach,istremis,patt,matt} zctyp;

const stein leer=0,kaiser=1,turm=2,geschuetz=3,pferd=4,
      elefant=5,bauer=6,mandarin=7;
const int mzugm=105;
const int xmax=8,ymax=9; //Anzahl Zeilen/Spalten-1
const zug leererzug={0,0,0,0};
const long punkte[3][8]={{0,7000,600,500,300,200,150,100},
		         {0,5000,580,480,280,180,180,80},
                         {0,9,6,5,3,2,2,1}};
const long unendl=2*punkte[0][kaiser];
const int hzmax=400;



class csbrettklasse : public CBoard {

  //Variablen
protected:
  int level; //aktuelles Level
  int alpha; //Alpha-Suchtiefe
  int alpha0;
  int beta; //Beta-Suchtiefe
  int beta0;
  int anztiefen; //Maximale Suchtiefe
  long **bew; //Bewertungsfeld
  zug **zi; //Feld der imaginren Zge
  int **ind; //Feld des Sortierungsindexes
  zug verbzug; //Verbotener Zug
  stein h[hzmax]; //Gespeicherte geschlagene Figuren
  int kaiserx[2];
  int kaisery[2];
  zug zr[hzmax]; //Feld der realen Zge
  stein stmoved[hzmax]; //field of moved stones

  bool lokkaiser(const int,int&,int&); //inline
  /* Diese Funktion gibt die x, und y Koordinate des Kaisers vom
     angegebenem Spieler zurck*/


  bool remis();
  /* Diese Funkton gibt wahr zurck, falls eine Remis-Situation
     eingetreten ist*/

  bool allgmoeglich(const zug&);
  /* Diese Funktion berprft ob ein Zug allgemein mglich ist,
     sie berprft ob
     1. auf dem Zielfeld sich ein Stein der selben Farbe wie
        auf dem Quellfeld befindet
     2. Die Kaiser auf einer Linie stehen ohne da sich ein
        Stein dazwischen befindet*/

  bool zuegevonfeld
        (const int& x,const int& y,const int& t,int &k,long *b0,
         bool schrbew,bool vollesuche,int *ix);
  /* Diese Funktion speichert alle mglichen Zge, die von
     Feld x,y aus mglich sind. Sonstige Parameter:
     t: Position an der die Daten gespeichert werden sollen
     k: Rckgabe: Anzahl Zge
     b0: Optimale Bewertungen, wenn gesetzt wird die Zugtiefe evtl.
         abgebrochen,nicht gesetzt:d.h.=NULL
     schrbew: Wenn wahr schreibt der Computer auch die zugehrigen
              Bewertungen in ein Feld
     vollesuche: wahr: Alle mglichen Zge werden zurckgegeben
                 falsch: Nur bestimmte Zge werden zurckgegeben
*/


  int zuegevonspieler
       (const int& sp,long* b0,bool schrbew,bool vollesuche,
        const int&t,int *ix);
/* Diese Funktion ermittelt alle mglichen Zge von Spieler sp
b0,schrbew,vollesuche,t s.o. Rckgabewert: Anzahl Zugmglichkeiten*/

  /* liefert die Bewertung eines Zuges zurck
     2. Paramter Zugtiefe*/
  long bewerte(const zug& z,const int& t);

  /** tries to avoid moving the same stone over and over*/
  long avoid_repetition(stein s);

  void aktverbzug();
  /* Diese Funktion aktualisiert den verbotenen Zug, d.h. analysiert
  ob 3 mal durch jeweils gleiche Zuege der gleiche Spielstand
  entstanden ist und ermittelt dann den Zug, der nicht ausgefhrt
  werden darf. Drfen alle Zuege ausgefhrt werden, speicher diese
  Funktion in verbzug den leeren Zug.*/

  bool aufnehmen
   (const int&t, int& k,const zug&z,
    const bool& alle,const bool&tubew,long* b0,int *ix);
  /* Speichert einen Zug in ein Feld ab
     t=Tiefe
     k=Anzahl abgespeicherter bis her
     alle=wahr alle Zge werden abgespeichert, sonst nur ausgewhlte
     tubew=wahr: schreibt auch Bewertungen in Feld*/

  stein ziehe(const zug&); //inline
  /* fhrt angegebenen Zug aus. Gibt Figur zurck, die auf dem
     Zielfeld steht,0 entspricht keine Figur*/

  void zuruecknehmen(const zug&,const stein&); //inline
  /* Nimmt Zug zurck. Ntige Parameter:
     1. Zug der zurckgenommen werden soll
     2. Stein, die evtl in diesem Zug geschlagen wurde*/

public:
        stein f[xmax+1][ymax+1]; //Spielfeld
	zug holezug(int i);
  int hz; //Anzahl der Halbzge
  
  csbrettklasse()
  {
	 zi=NULL;
	 ind=NULL;
	 bew=NULL; 	 
  }
  //Konstruktor v=Spielstrke

  ~csbrettklasse();


  void init(); //inline
  /* Initialisiert das Schachbrett auf Ausgangszustand
     Wird nicht automatisch vom Konstruktor aufgerufen*/

  
  bool ziehereg(const zug& z);
  /* Ziehe angegebenen Zug und speichere ihn
     Von auen sollte nur diese Funktion ausgefhrt werden*/


  stein zuruecknehmenreg();
  /* Nimmt gespeicherten Zug zurck und liefert den bei diesem Zug
  gespeicherten Stein zurueck*/

  bool moeglich(const zug&,const int&); //inline
  /* Gibt zurck, ob ein Zug von einem Spieler mglich ist*/

  bool moeglichvonfigur(const zug&);
  /* Gibt zuck ob ein bestimmter Stein von seiner Bewegungsart
     her diesen Zug ausfhren knnte*/

  zctyp analyse(const int&);
  /* analysiert das Spielbrett*/

  zug besterzug(int sp,int& command,int lev=0);
  /* Der Computer liefert den Zug zurck, den er am besten findet.
     Sp= Spieler am Zug
     abbruch=Flag das indiziert ob der Computer die Suche abbrechen
     soll. Gedacht fr Multithreading*/

  bool setzelevel(int v);
  /* Aendert die Spielstrke des Computers
     Parameter: v=neue Spielstrke
                altv=alte Spielstrke
  Achtung sollte diese Aktion schiefgehen ist das Objekt in
  einem undefinierten Zustand. Es sollte anschlieend der
  gleiche Befehl mit dem alten Level aufgerufen werden, wenn
  das Programm fortgesetzt werden soll.
  */
  
  bool schach(const int&);
  /* Diese Funktion gibt wahr zurck wenn der angegebene Spieler
     sich im Schach befindet, sondt falsch*/

  int getcontents(int x,int y)
  {
	  return (int)f[x][y];
  }
  
  void setzebrett(stein feld2[xmax+1][ymax+1]);
  // ersetzt das aktuelle Spielfeld durch feld2

  friend ostream& operator<<(ostream& os,const csbrettklasse& brett);
  friend istream& operator>>(istream& is,csbrettklasse& brett);
  
	
};
/*
inline int sgn(int i)
{
  return int((i>0)-(i<0));
}*/


//Operationen auf Zug

extern ostream& operator<<(ostream& os,const zug& z);
extern istream& operator>>(istream& is,zug& z);

//*********************** == ******************************
inline bool operator==(const zug& z1,const zug& z2) {
return (z1.x1==z2.x1 && z1.y1==z2.y1 &&
        z1.x2==z2.x2 && z1.y2==z2.y2);
}

//*********************** != ******************************
inline bool operator!=(const zug& z1,const zug& z2) {
return !(z1==z2);
}

/*************** ordnen ************************/
inline void ordnen(const int& k,int* ind,long* bew)
/* k=Anzahl zu ordnende Objekte
   ind*=Indexfeld das geordnet wird
   bew*=Objekte ber die die Ordnung bestimmt wird*/
{
  int h;
  register int i,i2;

  //Index initialisieren
  for (i=0;i<=k;i++)
    ind[i]=i;

  for (i=0;i<k;i++)
    for (i2=i+1;i2<k;i2++)
      if (bew[ind[i]]<bew[ind[i2]])
      {
	h=ind[i];
	ind[i]=ind[i2];
	ind[i2]=h;
      }

}

//********************** ZIEHE *****************************
inline stein csbrettklasse::ziehe(const zug& z){
stein quelle=f[z.x1][z.y1],ziel=f[z.x2][z.y2];

//Kaiser auf Quellfeld
if (abs(quelle)==kaiser)
{
  kaiserx[spvonstein(quelle)]=z.x2;
  kaisery[spvonstein(quelle)]=z.y2;
}
//Kaiser auf Zielfeld
else if (abs(ziel)==kaiser)
{
  kaiserx[spvonstein(ziel)]=-1; //Kaiser nicht mehr auf Brett
  kaisery[spvonstein(ziel)]=-1;
}
f[z.x2][z.y2]=quelle;//Zielfeld=Quellfeld
f[z.x1][z.y1]=leer; //Quellfeld lschen
return ziel; //gesicherten Wert zurckgeben
}

//*********************** ZURUECKNEHMEN ***********************
//nimmt Zug z zurck und besetzt evl. vorher geschlagenen Stein
// mit h
inline void csbrettklasse::zuruecknehmen(const zug& z,const stein& h){
stein quelle=f[z.x2][z.y2];
if (abs(h)==kaiser) //Kaiser war bei diesem Zug geschlagen worden
{
  kaiserx[spvonstein(h)]=z.x2; //da kommt der Kaiser hin
  kaisery[spvonstein(h)]=z.y2;
}
else if (abs(quelle)==kaiser)
{ //Kaiser wird zurckgesetzt
  kaiserx[spvonstein(quelle)]=z.x1;
  kaisery[spvonstein(quelle)]=z.y1;
}
f[z.x1][z.y1]=quelle; //Stein zurckziehen
f[z.x2][z.y2]=h; //evtl. Stein zurck aufs Brett stellen
}

//************************ MOEGLICH ***************************
// gibt zurck ob Zug z von Spieler sp mglich ist
inline bool csbrettklasse::moeglich(const zug& z,const int& sp){
/* sp: Spieler fr den Zug berprft werden soll
   z: Zug, der berprft werden soll*/
bool imschach;
stein h;

if (z==verbzug) return false;
if (!allgmoeglich(z) || (f[z.x1][z.y1]<0)!=sp) return false;
h=ziehe(z);
imschach=schach(sp);
zuruecknehmen(z,h);
return (!imschach && moeglichvonfigur(z));
}



//************************* BEWERTE ***************************
inline long csbrettklasse::bewerte(const zug&z,const int& t) {
long bew,b2=0;
stein h=abs(f[z.x2][z.y2]),st=f[z.x1][z.y1];
int kx=kaiserx[!spvonstein(st)],ky=kaisery[!spvonstein(st)];

//Materialstrategie

bew=(long)(punkte[0][h]*pow(0.8,t));

// Verhindern, da Computer falsch auf Schach reagiert
if (h==kaiser && t==1) bew=bew+bew+bew;

//Bewegungsfreiheit Kaiser
if (st==kaiser && (z.x1==3 || z.x1==5) && z.x2==4)
{
  for (int y=0;y<=2;y++)
  {
     if (f[z.x1][y]!=leer) bew+=2;
  }
}
else if (st==-kaiser)
{
  for (int y=ymax-2;y<=ymax;y++)
  {
    if (f[z.x1][y]!=leer) bew+=2;
  }
}

// Angriffsstrategie
if (abs(st)==turm || abs(st)==geschuetz || abs(st)==pferd ||
    abs(st)==bauer)
    bew=bew+10-abs(z.x2-kx)+abs(z.x1-kx)
              -abs(z.y2-ky)+abs(z.y1-ky);

if (abs(st)==turm) bew++;

//avoid repetition
if (t==0) bew=bew-avoid_repetition(st);

if (bew<0) return 0;
return bew;
}

/** avoid moving the same piece to much*/
inline long csbrettklasse::avoid_repetition(stein movingStone)
{
  int i=0,hznr=hz;
  float b=0;
  int stone=abs(movingStone);
  if (stone!=turm && stone!=geschuetz && stone!=pferd)
    return 0;
  
  while (hznr>=2 && i<5)
  {
    hznr=hznr-2;
    i++;
    if (stmoved[hznr]==movingStone) b=1.4*b+1.5;
  }
  return (long) b;
}


/************************ ZUG AUFNEHMEN ***********************/
inline bool csbrettklasse::aufnehmen
(const int& t,int& k,const zug& z,const bool& alle,const bool& tubew,
 long* b0,int *ix)
/* t=Tiefe
   k=Anzahl aufgenomener Zge
   alle=alle aufnehmen
   tubew=Bewertungen auch abspeichern
   Gibt bei Zugoptimierung falsch zurueck
*/
{


  //Nur Aufnahme, wenn Figuren geschlagen werden?
  if (!alle && f[z.x2][z.y2]==0) return true;

  //Zug ist verbotener Zug?
  if (t==0 && z==verbzug) return true;
  
  zi[t][k]=z;
  if (tubew) bew[t][k]=bewerte(z,t);


  /*Mensch-Optimierung
    Der Zug, den der Mensch im letzten Zug macht ist so schlecht
    fr den Computer, da er auf keinen Fall den dafr
    zugrundeliegenden Zug ausfhren wird*/

  if ((t==beta || t==alpha && alpha>1) && b0)
  {
    if (bew[t-1][ind[t-1][ix[t-1]]]-bew[t][k]<b0[t-1])
    {
      if (t==beta || (t==alpha && b0[t-1]>0)) return false;
    }
  }

  k++;
  return true;
}


//************************** ZUEGEVONFELD **********************
inline bool csbrettklasse::zuegevonfeld
(const int&x,const int&y,const int& t,int& k,long *b0,
bool schrbew,bool vollesuche,int *ix){
#define zugaufnehmen if (!aufnehmen(t,k,z,vollesuche,schrbew,b0,ix)) \
                     return false;

zug z={x,y,0,0};
int i;
register int dx,dy;
bool sprung;
const static int dxa[]={0,1,0,-1},dya[]={1,0,-1,0};

switch (abs(f[x][y])) {
case kaiser:
  for (i=0;i<=3 ;i++ ) {
    z.x2=z.x1+dxa[i];
    z.y2=z.y1+dya[i];
    if (impalast(z.x2,z.y2) && allgmoeglich(z)) zugaufnehmen;
  } /* endfor */
  break;
case mandarin:
  for (dx=-1;dx<=1 ;dx+=2 ) {
    for (dy=-1;dy<=1 ;dy+=2 ) {
      z.x2=z.x1+dx;
      z.y2=z.y1+dy;
      if (impalast(z.x2,z.y2) && allgmoeglich(z)) zugaufnehmen;
    } /* endfor */
  } /* endfor */
  break;
case turm:
  for (i=0;i<=3 ;i++ ) {
    z.x2=z.x1;
    z.y2=z.y1;
    do {
      z.x2=z.x2+dxa[i];
      z.y2=z.y2+dya[i];
      if (imfeld(z.x2,z.y2)) {
        if (allgmoeglich(z)) zugaufnehmen;
      } else break; //auerhalb des Spielfeldes
    } while (f[z.x2][z.y2]==leer  ); /* enddo */
  } /* endfor */
  break;
case geschuetz:
  
  for (i=0;i<=3 ;i++ ) {
    z.x2=z.x1;
    z.y2=z.y1;
    for (;;)
    {
      z.x2=z.x2+dxa[i];
      z.y2=z.y2+dya[i];
      if (!imfeld(z.x2,z.y2)) break;
      if (!allgmoeglich(z)) break;
      if (f[z.x2][z.y2]==leer)
      {
	zugaufnehmen;
	continue;
      }
      else
      {
	for (;;)
	{
	  z.x2=z.x2+dxa[i];
	  z.y2=z.y2+dya[i];
	  if (!imfeld(z.x2,z.y2)) break; //not inside the field?
	  if (f[z.x2][z.y2]==leer) continue; //target position empty
	  if (allgmoeglich(z)) zugaufnehmen;
	  break;
	}
      }
      break;
    } /* end for(;;) */
  } /* endfor */
  break;
case pferd:
  for (dx=-2;dx<=2 ;dx++ ) {
    if (dx==0) continue;
    dy=3-abs(dx);
    for (i=1;i<=2 ;i++,dy=-dy ) {
      z.x2=z.x1+dx; //Alle Pferdpositionen
      z.y2=z.y1+dy;
      if (f[z.x2-sgn(dx)][z.y2-sgn(dy)]==leer) {
        if (imfeld(z.x2,z.y2) && allgmoeglich(z)) zugaufnehmen;
      } /* endif */
    } /* endfor */
  } /* endfor */
  break;
case elefant:
  for (dx=-1;dx<=1 ;dx+=2 ) {
    for (dy=-1;dy<=1 ;dy+=2 ) {
      z.x2=z.x1+dx+dx;
      z.y2=z.y1+dy+dy;
      if (imfeld(z.x2,z.y2) && //innerhalb Spielfeld
         (sgn(f[z.x1][z.y1])>0 && z.y2<=4 || //auf eigener
          sgn(f[z.x1][z.y1])<0 && z.y2>4) && //Fluseite
          f[z.x1+dx][z.y1+dy]==leer && //Zwischenfeld leer?
          allgmoeglich(z))  //Zug allgemein mglich?
            zugaufnehmen;
    } /* endfor */
  } /* endfor */
  break;
case bauer:
  z.x2=z.x1;                        //Bauer von vorne
  z.y2=z.y1+sgn(f[z.x1][z.y1]);
  if (0<=z.y2 && z.y2<=ymax && allgmoeglich(z)) zugaufnehmen;
  if (sgn(f[z.x1][z.y1])==1  && z.y1>4 ||
      sgn(f[z.x1][z.y1])==-1 && z.y1<=4) { //andere Fluseite
    z.y2=z.y1;
    if (z.x1<xmax) {   //Bauer nach rechts
      z.x2=z.x1+1;
      if (allgmoeglich(z)) zugaufnehmen;
    } /* endif */
    if (z.x1>0) {   //Bauer nach links
      z.x2=z.x1-1;
      if (allgmoeglich(z)) zugaufnehmen;
    } /* endif */
  } /* endif */
} /* endswitch */
return true;
}
#endif












